// Minecraft Standalone Client
// Copyright (C) 2009-2010 Colin Bartolome
// Licensed under the GPL. See LICENSE.txt for details.

import java.awt.event.*;
import java.io.*;
import java.util.Enumeration;
import java.util.zip.*;

import javax.swing.*;

public class LoginDialog extends JDialog
{
   private static class TexturePack
   {
      private File file;
      
      private TexturePack(File file)
      {
         this.file = file;
      }
      
      @Override
      public String toString()
      {
         return file == null ? "<default>" : file.getName();
      }
   }
   
   private static final TexturePack DEFAULT_TEXTURE_PACK = new TexturePack(null);
   
   public LoginDialog(final Main main)
   {
      super(main, "Login", true);
      
      final String texturePack = main.getTexturePack();
      String loginCookie = main.getLoginCookie();
      int pipeIndex = loginCookie.indexOf('|');
      String username;
      String password;
      
      if (pipeIndex == -1)
      {
         username = password = "";
      }
      else
      {
         username = loginCookie.substring(0, pipeIndex);
         password = loginCookie.substring(pipeIndex + 1);
      }
      
      JPanel panel = new JPanel();
      final JTextField usernameField;
      final JPasswordField passwordField;
      final JCheckBox checkBox;
      final JComboBox comboBox;
      JButton button;
      Object[] texturePacks = getTexturePacks();
            
      panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
      panel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
      
      panel.add(new JLabel("Note: Password is stored in plain text."));

      panel.add(Box.createVerticalStrut(5));
      
      Box box;
      
      box = Box.createHorizontalBox();
      box.add(new JLabel("Username:"));
      box.add(Box.createHorizontalStrut(5));
      box.add(usernameField = new JTextField(10));
      panel.add(box);
      
      usernameField.setText(username);
      
      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(new JLabel("Password:"));
      box.add(Box.createHorizontalStrut(5));
      box.add(passwordField = new JPasswordField(10));
      panel.add(box);
      
      passwordField.setText(password);
      
      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(Box.createHorizontalGlue());
      box.add(checkBox = new JCheckBox());
      box.add(Box.createHorizontalStrut(5));
      box.add(new JLabel("Save Information"));
      box.add(Box.createHorizontalGlue());
      panel.add(box);
      
      checkBox.setSelected(!username.isEmpty() || !password.isEmpty());
      
      panel.add(Box.createVerticalStrut(5));
      
      box = Box.createHorizontalBox();
      box.add(Box.createHorizontalGlue());
      box.add(new JLabel("Textures:"));
      box.add(Box.createHorizontalStrut(5));
      box.add(comboBox = new JComboBox(texturePacks));
      box.add(Box.createHorizontalGlue());
      panel.add(box);
      
      if (texturePack != null)
      {
         for (int i = 1; i < texturePacks.length; i++)
         {
            if (texturePacks[i].toString().equals(texturePack))
            {
               comboBox.setSelectedIndex(i);
               break;
            }
         }
      }

      panel.add(Box.createVerticalStrut(5));

      box = Box.createHorizontalBox();
      box.add(Box.createHorizontalGlue());
      box.add(button = new JButton("Go!"));
      box.add(Box.createHorizontalGlue());
      panel.add(box);
      
      button.addActionListener(new ActionListener()
      {
         @SuppressWarnings("deprecation")
         @Override
         public void actionPerformed(ActionEvent evt)
         {
            if (comboBox.getSelectedIndex() == 0)
               main.setTexturePack(null);
            else
               main.setTexturePack(comboBox.getSelectedItem().toString());
            
            main.setLoginCookie(usernameField.getText().trim(), passwordField.getText().trim(), checkBox.isSelected());
            
            setTexturePack((TexturePack)comboBox.getSelectedItem());
            setVisible(false);
         }
      });
      
      getRootPane().setDefaultButton(button);
      
      add(panel);
      
      pack();
      setLocationRelativeTo(null);
      setVisible(true);
   }
   
   private static Object[] getTexturePacks()
   {
      File[] files;
      
      try
      {
         File file = new File(LoginDialog.class.getResource("/texturePacks/").toURI());
         
         files = file.listFiles(new FilenameFilter()
         {
            @Override
            public boolean accept(File dir, String name)
            {
               return name.toLowerCase().endsWith(".zip");
            }
         });
      }
      catch (Exception e)
      {
         System.err.println("Could not read list of texture packs: " + e.toString());
         
         files = new File[0];
      }
      
      Object[] texturePacks = new Object[files.length + 1];
      
      texturePacks[0] = DEFAULT_TEXTURE_PACK;
      
      for (int i = 0; i < files.length; i++)
         texturePacks[i + 1] = new TexturePack(files[i]);
   
      return texturePacks;
   }
   
   private static void setTexturePack(TexturePack texturePack)
   {
      try
      {
         File dir = new File("textures/");
         
         deleteDirectory(dir);
         
         if (texturePack.file != null)
         {
            dir.mkdir();
            unzip(texturePack.file, dir);
         }
      }
      catch (Exception e)
      {
         System.err.println("Failed to change texture pack: " + e.toString());
      }
   }
   
   private static void deleteDirectory(File dir)
   {
      if (!dir.exists())
         return;
      
      for (File file : dir.listFiles())
      {
         if (file.isDirectory())
           deleteDirectory(file);
         else
           file.delete();
      }
   }
   
   private static void unzip(File file, File dest) throws IOException
   {
      ZipFile zipFile = new ZipFile(file);
      Enumeration<? extends ZipEntry> entries = zipFile.entries();
      
      while (entries.hasMoreElements())
      {
         ZipEntry entry = entries.nextElement();
         
         if (!entry.isDirectory())
         {
            File newFile = new File(dest, entry.getName());
            
            createParentDirectories(newFile);
            extractFile(zipFile, entry, newFile);
         }
      }
   }
   
   private static void createParentDirectories(File file)
   {
      File parent = file.getParentFile();
      
      if (parent != null)
         parent.mkdirs();
   }
   
   private static void extractFile(ZipFile zipFile, ZipEntry entry, File file) throws IOException
   {
      int bytesRead;
      byte[] buffer = new byte[1024];
      BufferedInputStream in = new BufferedInputStream(zipFile.getInputStream(entry));
      BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(file));
      
      while ((bytesRead = in.read(buffer)) != -1)
         out.write(buffer, 0, bytesRead);
      
      in.close();
      out.close();
   }
}
