// Minecraft Standalone Client
// Copyright (C) 2009-2010 Colin Bartolome
// Licensed under the GPL. See LICENSE.txt for details.

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.util.TimerTask;

import javax.swing.*;

import com.mojang.minecraft.Entity;
import com.mojang.minecraft.level.Level;
import com.mojang.minecraft.net.NetworkPlayer;

public class MapPanel extends JPanel
{
   private enum MapType
   {
      NORMAL("Normal"),
      SOLID("Solid"),
      ELEVATION("Height"),
      RADAR("Radar"),
      SLICE("Slice");
      
      private String name;
      
      private MapType(String name)
      {
         this.name = name;
      }
      
      @Override
      public String toString()
      {
         return name;
      }
      
      private MapType getNext()
      {
         int ordinal = ordinal() + 1;
         
         if (ordinal >= values().length)
            ordinal = 0;
         
         return values()[ordinal];
      }
      
      private MapType getPrevious()
      {
         int ordinal = ordinal() - 1;
         
         if (ordinal < 0)
            ordinal = values().length - 1;
         
         return values()[ordinal];
      }
      
      private int getHighest(Level level, int x, int z)
      {
         TileType[] tileTypes = TileType.values();
         
         for (int y = level.depth - 1; y > 0; y--)
         {
            TileType tileType = tileTypes[level.getTile(x, y, z)];
         
            if (tileType == TileType.AIR || tileType == TileType.GLASS
                  || ((this == SOLID || this == ELEVATION) && (tileType == TileType.INFINITE_WATER || tileType == TileType.FINITE_WATER
                     || tileType == TileType.INFINITE_LAVA || tileType == TileType.FINITE_LAVA)))
               continue;

            return y;
         }
         
         return 0;
      }
      
      private Color getColor(Level level, int x, int playerY, int z)
      {
         float brightness;
         
         switch (this)
         {
            case NORMAL:
            case SOLID:
               return TileType.values()[level.getTile(x, getHighest(level, x, z), z)].getColor();
            case ELEVATION:
               brightness = (float)getHighest(level, x, z) / level.depth;
               
               return new Color(brightness, brightness, brightness);                  
            case RADAR:
               int air = 0;
               
               for (int y = 0; y < playerY; y++)
                  if (TileType.values()[level.getTile(x, y, z)] == TileType.AIR)
                     air++;

               brightness = (float)air / playerY;
               
               return new Color(brightness, brightness, brightness);
            case SLICE:
               return TileType.values()[level.getTile(x, playerY, z)].getColor();
         }
         
         return null;
      }
   }
   
   private static final long PERIOD_REPAINT = 200;
   
   private static final Color COLOR_ENTITY = new Color(128, 64, 64);
   
   private static final Color COLOR_PLAYER = Color.RED;
   
   private static final Color COLOR_EYE = Color.WHITE;
   
   private static final int DEGREES = 360;
   
   private static final int ZOOM_THRESHHOLD = 256;
   
   private static final String FORMAT_TITLE = "Mode: %s   Zoom: %d";
   
   private static final int MIN_ENTITY_SIZE = 4;
   
   private static final int MIN_EYE_SIZE = MIN_ENTITY_SIZE / 2;
   
   private JFrame frame;
   
   private CustomPlayer player;
   
   private BufferedImage image;
   
   private Font font;
   
   private MapType mapType = MapType.NORMAL;
   
   private Level oldLevel;
   
   private java.util.Timer repaintTimer;
   
   private boolean updatingMap;
   
   private boolean resetMap;
   
   private int zoom;
   
   private int entitySize;
   
   private int eyeSize;
   
   MapPanel(final JFrame frame, CustomPlayer player, Level level)
   {
      this.frame = frame;
      this.player = player;
      oldLevel = level;
      zoom = getDefaultZoom(level);
      
      Thread mapThread = new Thread(new Runnable()
      {
         @Override
         public void run()
         {
            updateMap();
         }
      }, "Map Updater");
      
      mapThread.setDaemon(true);
      mapThread.start();

      (repaintTimer = new java.util.Timer("Repaint", true)).schedule(new TimerTask()
      {
         @Override
         public void run()
         {
            repaint();
         }
      }, 0, PERIOD_REPAINT);
      
      addMouseListener(new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent evt)
         {
            mapType = evt.getButton() == MouseEvent.BUTTON1 ? mapType.getNext() : mapType.getPrevious();
            
            updateTitle();
         }
      });
      
      addMouseWheelListener(new MouseWheelListener()
      {
         @Override
         public void mouseWheelMoved(MouseWheelEvent evt)
         {
            if (evt.getWheelRotation() < 0)
            {
               zoom++;
               resize(false);
            }
            else if (zoom > 1)
            {
               zoom--;
               resize(false);
            }
         }
      });
      
      resize(false);
   }
   
   @Override
   protected void paintComponent(Graphics g)
   {
      Level level = player.getLevel();
      
      if (level != oldLevel)
      {
         oldLevel = level;
         resize(true);
      }
      
      BufferedImage localImage = image;
      Color oldColor = g.getColor();
      
      if (font == null)
      {
         font = g.getFont().deriveFont(Font.BOLD);
      }
      
      g.setFont(font);
      
      if (localImage == null)
      {
         g.setColor(Color.BLACK);
         g.fillRect(0, 0, getWidth(), getHeight());
         
         g.setColor(Color.WHITE);
         g.drawString("Updating map...", 10, getHeight() / 2);
      }
      else
      {
         g.drawImage(localImage, 0, 0, this);
      }
      
      for (Object object : player.getLevel().blockMap.all)
      {
         Entity entity = (Entity)object;
         boolean drawExtra = entity instanceof NetworkPlayer;
         
         g.setColor(COLOR_ENTITY);
         drawEntity(g, entity.x, entity.z, entity.yRot, drawExtra, true);
         
         if (drawExtra)
         {
            drawEntityName(g, ((NetworkPlayer)entity).displayName, entity.x, entity.z);
         }
      }
      
      g.setColor(COLOR_PLAYER);
      drawEntity(g, player.x, player.z, player.yRot, true, false);
      drawEntityName(g, "You", player.x, player.z);

      g.setColor(oldColor);
   }
   
   private void drawEntity(Graphics g, float x, float z, float yRot, boolean drawEye, boolean reverseHeading)
   {
      int drawX = (int)(x * zoom - zoom / 2);
      int drawY = (int)(z * zoom - zoom / 2);
      
      g.fillRect(drawX, drawY, entitySize, entitySize);
      
      if (drawEye)
      {
         int heading = (int)yRot % DEGREES;
         
         if (heading < 0)
            heading += DEGREES;
         
         if (reverseHeading)
            heading = DEGREES - heading;
         
         switch (heading / (DEGREES / 12))
         {
            case 0:
               drawX += eyeSize;
               drawY -= eyeSize;
               break;
            case 1:
               drawX += eyeSize * 2;
               drawY -= eyeSize;
               break;
            case 2:
               drawX += eyeSize * 2;
               break;
            case 3:
               drawX += eyeSize * 2;
               drawY += eyeSize;
               break;
            case 4:
               drawX += eyeSize * 2;
               drawY += eyeSize * 2;
               break;
            case 5:
               drawX += eyeSize;
               drawY += eyeSize * 2;
               break;
            case 6:
               drawY += eyeSize * 2;
               break;
            case 7:
               drawX -= eyeSize;
               drawY += eyeSize * 2;
               break;
            case 8:
               drawX -= eyeSize;
               drawY += eyeSize;
               break;
            case 9:
               drawX -= eyeSize;
               break;
            case 10:
               drawX -= eyeSize;
               drawY -= eyeSize;
               break;
            case 11:
               drawY -= eyeSize;
               break;
         }
         
         g.setColor(COLOR_EYE);
         
         g.fillRect(drawX, drawY, eyeSize, eyeSize);
      }
   }
   
   private void drawEntityName(Graphics g, String name, float x, float z)
   {
      g.setColor(Color.BLACK);
      g.drawString(name, (int)(x * zoom) + entitySize + 3 + 1, (int)(z * zoom) + entitySize + 1);
      g.setColor(Color.WHITE);
      g.drawString(name, (int)(x * zoom) + entitySize + 3, (int)(z * zoom) + entitySize);
   }
   
   private void updateMap()
   {
      MapType localMapType = null;
      Graphics g = null;

      updatingMap = true;
      
      while (updatingMap)
      {
         Level level = player.getLevel();
         
         if (localMapType != mapType || resetMap)
         {
            localMapType = mapType;
            resetMap = false;
            
            if (g != null)
               g.dispose();
            
            image = new BufferedImage(level.width * zoom, level.height * zoom, BufferedImage.TYPE_INT_RGB);
            g = image.createGraphics();

            g.setColor(TileType.ADMIN.getColor());
            g.fillRect(0, 0, getWidth(), getHeight());
         }

         int playerY = (int)player.y - 1;
         
         for (int x = 0; x < level.width; x++)
         {
            for (int z = 0; z < level.height; z++)
            {
               g.setColor(localMapType.getColor(level, x, playerY, z));
               
               g.fillRect(x * zoom, z * zoom, zoom, zoom);
   
               Thread.yield();
            }
   
            if (localMapType != mapType || resetMap)
               break;
         }
      }
      
      if (g != null)
         g.dispose();
   }

   void stop()
   {
      updatingMap = false;
      
      if (repaintTimer != null)
         repaintTimer.cancel();
   }

   private int getDefaultZoom(Level level)
   {
      return level.width > ZOOM_THRESHHOLD || level.height > ZOOM_THRESHHOLD ? 1 : 2;
   }
   
   private void resize(boolean rezoom)
   {
      if (rezoom)
         zoom = getDefaultZoom(oldLevel);
      
      entitySize = zoom * 2;
      if (entitySize < MIN_ENTITY_SIZE)
         entitySize = MIN_ENTITY_SIZE;
      
      eyeSize = entitySize / 2;
      if (eyeSize < MIN_EYE_SIZE)
         eyeSize = MIN_EYE_SIZE;

      Dimension size = new Dimension(oldLevel.width * zoom, oldLevel.height * zoom);
   
      setMinimumSize(size);
      setPreferredSize(size);
      frame.pack();
      
      image = null;
      resetMap = true;
      
      updateTitle();
   }
   
   private void updateTitle()
   {
      frame.setTitle(String.format(FORMAT_TITLE, mapType.toString(), zoom));
   }
}
